unit unzipfile;

interface

uses Windows, zlibpas;

const
  ZIP_FILE_HEADER_SIGN = 'PK'#03#04;
  ZIP_EXTENDED_HEADER_SIGN = 'PK'#07#08;
  ZIP_CENTRALDIR_HEADER_SIGN = 'PK'#01#02;
  ZIP_END_CENTRALDIR_HEADER_SIGN = 'PK'#05#06;

  UNZIP_OK = 0;
  UNZIP_MEMERR = 1;
  UNZIP_BROKEN = 2;

type
  PZipFileHeader = ^TZipFileHeader;
  TZipFileHeader = packed record
    VersionNeeded: Word;
    Flag: Word;
    Method: Word;
    FileMTime: Word;
    FileMDate: Word;
    Crc: Longint;
    CompSize: Longint;
    UnCompSize: Longint;
    FileNameLen: Word;
    ExtraFieldLen: Word;
  end;

  TZipExtendedHeader = packed record
    Crc: Longint;
    CompSize: Longint;
    UnCompSize: Longint;
  end;

  TCentralDirectory = packed record
    VersionMadeBy: Word;
    VersionNeeded: Word;
    Flag: Word;
    Method: Word;
    FileMTime: Word;
    FileMDate: Word;
    Crc: Longint;
    CompSize: Longint;
    UnCompSize: Longint;
    FileNameLen: Word;
    ExtraFieldLen: Word;
    FileCommentLen: Word;
    DiskNumberStart: Word;
    InternalFileAttr: Word;
    ExternalFileAttr: Longint;
    LocalHeaderOffset: Longint;
  end;

  TEndOfCentralDirectory = packed record
    NumThisDisk: Word;
    NumDiskStartDir: Word;
    NumEntriesCentralDirThisDisk: Word;
    TotalEntriesCentralDir: Word;
    SizeCentralDir: Longint;
    OffsetStartCentralDir: Longint;
    ZipFileCommentLen: Word;
  end;

  PSign = ^TSign;
  TSign = array[0..3] of Char;

function ReadEndOfCentralDir(HFile: THandle; var ECDir: TEndOfCentralDirectory;
  var Offset: Integer): Boolean;
function ReadCentralDir(HFile: THandle; var CDir: TCentralDirectory; 
  var FileName: string): Boolean;
function ReadZipFileHeader(HFile: THandle; var Header: TZipFileHeader;
  var FileName: string): Boolean;
function Extract(HFile: THandle; Header: TZipFileHeader; Buf: PChar; 
  BufSize: Cardinal): Integer;

implementation

// End of Cental DirǂŃt@C|C^Central Dir̐擪ɃZbg
// (t@C64KB܂ł̂ڂEnd of Cental DirT)
function ReadEndOfCentralDir(HFile: THandle; var ECDir: TEndOfCentralDirectory;
  var Offset: Integer): Boolean;
var
  sign: TSign;
  readbytes: DWORD;
  totalread: Integer;
  readbuf: string;
  idx: Integer;
begin
  SetFilePointer(HFile, -(SizeOf(sign)+SizeOf(TEndOfCentralDirectory)), nil, FILE_END);
  ReadFile(HFile, sign, SizeOf(sign), readbytes, nil);
  if readbytes < SizeOf(sign) then
  begin
    Result := False;
    Exit;
  end;

  if sign <> ZIP_END_CENTRALDIR_HEADER_SIGN then
  begin
    totalread := 0;
    SetLength(readbuf, 1024*8);
    while True do
    begin
      SetFilePointer(HFile, -Length(readbuf), nil, FILE_CURRENT);
      FillChar(readbuf[1], Length(readbuf), 0);
      ReadFile(HFile, readbuf[1], Length(readbuf), readbytes, nil);
      Inc(totalread, readbytes);
      idx := Pos(ZIP_END_CENTRALDIR_HEADER_SIGN, readbuf);
      if idx > 0 then
      begin
        SetFilePointer(HFile, -readbytes+idx-1+SizeOf(sign), nil, FILE_CURRENT);
        Break;
      end;
  
      if (readbytes < DWORD(Length(readbuf))) or (totalread > 65535) then
      begin
        Result := False;
        Exit;
      end;
  
      // ESignatureꍇl4bytesԂ点ēǂ
      SetFilePointer(HFile, -Length(readbuf)+SizeOf(sign), nil, FILE_CURRENT);
    end;
  end;

  ReadFile(HFile, ECDir, SizeOf(ECDir), readbytes, nil);

  // End of Central DirɋL^ĂCentral Dir̃TCYvZ
  // Central DirJnʒu End of Central DirɋL^ĂJnʒu
  // Local FilẽItZbgɑ(ɃS~Ăꍇ̑΍)
  // (ԂunzipƓ)
  Offset := Integer(SetFilePointer(hfile, -(SizeOf(sign)+SizeOf(ECDir)+ecdir.SizeCentralDir),
    nil, FILE_CURRENT)) - ECDir.OffsetStartCentralDir;

  Result := True;
  Exit;
end;

// Central Directoryƃt@CǂŃt@C|C^
// Central DirectoryɃZbg
function ReadCentralDir(HFile: THandle; var CDir: TCentralDirectory; 
  var FileName: string): Boolean;
var
  sign: TSign;
  readbytes: DWORD;
begin
  ReadFile(HFile, sign, SizeOf(sign), readbytes, nil);
  if (readbytes < SizeOf(sign)) or (sign <> ZIP_CENTRALDIR_HEADER_SIGN) then
  begin
    Result := False;
    Exit;
  end;

  ReadFile(HFile, CDir, SizeOf(CDir), readbytes, nil);
  SetLength(FileName, CDir.FileNameLen);
  ReadFile(HFile, FileName[1], CDir.FileNameLen, readbytes, nil);
  SetFilePointer(HFile, CDir.ExtraFieldLen+CDir.FileCommentLen, nil, FILE_CURRENT);

  Result := True;
end;

function ReadZipFileHeader(HFile: THandle; var Header: TZipFileHeader;
  var FileName: string): Boolean;
var
  sign: TSign;
  readbytes: DWORD;
begin
  sign := '';
  ReadFile(HFile, sign, SizeOf(sign), readbytes, nil);
  if sign <> ZIP_FILE_HEADER_SIGN then
  begin
    Result := False;
    Exit;
  end;

  ReadFile(HFile, Header, SizeOf(Header), readbytes, nil);
  if readbytes < SizeOf(Header) then
  begin
    Result := False;
    Exit;
  end;
  SetLength(FileName, header.FileNameLen);
  ReadFile(HFile, FileName[1], header.FileNameLen, readbytes, nil);
  SetFilePointer(HFile, Header.ExtraFieldLen, nil, FILE_CURRENT);

  Result := True;
end;

function Decompress(InBuf: PChar; InBufSize: Cardinal;
  OutBuf: PChar; OutBufSize: Cardinal): Integer;
var
  zs: z_stream;
begin
  FillChar(zs, SizeOf(z_stream), 0);
  zs.next_in := InBuf;
  zs.avail_in := InBufSize;
  zs.next_out := OutBuf;
  zs.avail_out := OutBufSize;
  Result := inflateInit2(zs, -15);
  if Result < Z_OK then
    Exit;
  try
    Result := inflate(zs, Z_NO_FLUSH);
  finally
    inflateEnd(zs);
  end;
end;

function Extract(HFile: THandle; Header: TZipFileHeader; Buf: PChar; 
  BufSize: Cardinal): Integer;
var
  readbytes : DWORD;
  inbuf: PChar;
begin
  inbuf := nil;
  GetMem(inbuf, Header.CompSize);
  try
    if inbuf = nil then
    begin
      Result := UNZIP_MEMERR;
      Exit;
    end;

    // check error?
    ReadFile(HFile, inbuf^, Header.CompSize, readbytes, nil);

    case Header.Method of
      0: Move(inbuf^, Buf^, Header.UnCompSize); // Stored
      Z_DEFLATED:
        if Decompress(inbuf, Header.CompSize, Buf, Header.UnCompSize) < Z_OK then
        begin
          Result := UNZIP_BROKEN;
          Exit;
        end;
    else
      Result := UNZIP_BROKEN;
      Exit;
    end;

    Result := UNZIP_OK;
  finally
    FreeMem(inbuf);
  end;
end;

end.
